// zig fmt: off
const std = @import("std");
const c = @import("constants.zig");
const bytes = @import("bytes.zig");
const oaep = @import("oaep.zig");
const prime = @import("prime.zig");

pub const Keys = struct {
  pubKey: c.rsaOutBits = 0,
  e: u32 = c.e,
  privKey: c.rsaOutBits = 0,
};

pub fn keyGen() Keys {
  var out: Keys = Keys{};

  const p = prime.Gen();
  const q = prime.Gen();
  var temp: c.doubleRsaOutBits = p;
  temp *= q;
  out.pubKey = @truncate(temp);

  var i: c.rsaOutBits = 0;
  var d: c.rsaOutBits = 0;
  temp = (p - 1);
  temp *= (q - 1);

  while (true) {
    if (@mod((temp * i) + 1, c.e) == 0) {
      d = @truncate(@divExact((temp * i) + 1, c.e));
      break;
    } else {
      i += 1;
    }
  }

  out.privKey = d;

  return out;
}

pub fn encrypt(inp: []const u8, pubKey: c.rsaOutBits) oaep.PlusEncodeErr!c.rsaOutBits {
  const encoded = oaep.plusEncode(inp) catch |err| {
    return err;
  };

  const enc = bytes.toInt(c.rsaOutBits, (encoded)[0..]);
  return prime.powerMod2(enc, c.e, pubKey);
}

pub fn decrypt(inp: c.rsaOutBits, pubKey: c.rsaOutBits, privKey: c.rsaOutBits) oaep.PlusDecodeErr![c.maxMesLen]u8 {
  const enc = prime.powerMod2(inp, privKey, pubKey);
  var encoded = [_]u8{0x00} ** c.rsaOutBytes;
  bytes.fromInt(enc, (encoded)[0..]);

  const decoded = oaep.plusDecode(encoded) catch |err| {
    return err;
  };

  return decoded;
}

pub fn sign(inp: []const u8, privKey: c.rsaOutBits, pubKey: c.rsaOutBits) c.rsaOutBits {
  const hashed = bytes.toInt(c.rsaOutBits, bytes.hashBytes(inp)[0..]);
  return prime.powerMod2(hashed, privKey, pubKey);
}

pub fn verify(inp: c.rsaOutBits, message: []const u8, pubKey: c.rsaOutBits) bool {
  const hash = bytes.hashBytes(message);
  const hashed: [c.hashOutBytes]u8 = [_]u8{0x00} ** (c.hashOutBytes - c.hashOutBytes) ++ hash;
  const out_n = prime.powerMod2(inp, c.e, pubKey);
  var out = [_]u8{0x00} ** c.hashOutBytes;
  bytes.fromInt(out_n, (out)[0..]);

  for (0..c.hashOutBytes) |i| {
    if (hashed[i] != out[i]) return false;
  }

  return true;
}

test "gen key" {
  std.debug.print("鍵を作成中・・・\n", .{});
  const aliceKeys = keyGen();
  std.debug.print("アリスちゃんの鍵：\n公開鍵： {any}\n秘密鍵： {any}\n", .{ aliceKeys.pubKey, aliceKeys.privKey });
  std.debug.print("鍵を作成中・・・\n", .{});
  const tanakaKeys = keyGen();
  std.debug.print("田中くんの鍵：\n公開鍵： {any}\n秘密鍵： {any}\n", .{ tanakaKeys.pubKey, tanakaKeys.privKey });
  const aliceSign = sign("暗号", aliceKeys.privKey, aliceKeys.pubKey);
  std.debug.print("アリスちゃんの「暗号」向けの証明： {any}\n", .{aliceSign});
  const tanakaVerify = verify(aliceSign, "暗号", aliceKeys.pubKey);
  std.debug.print("田中くんは証明を確認する： {any}\n", .{tanakaVerify});
  const aliceToTanakaEncrypted = encrypt("こんにちは", tanakaKeys.pubKey);
  std.debug.print("アリスちゃんは田中くんに「こんにちは」という暗号されたメッセージを送った： {any}\n", .{aliceToTanakaEncrypted});
  const aliceToTanakaDecrypted = decrypt(try aliceToTanakaEncrypted, tanakaKeys.pubKey, tanakaKeys.privKey);
  std.debug.print("田中くんはメッセージを復号された： {any}\n", .{aliceToTanakaDecrypted});
}

test "encrypt-decrypt round-trip" {
  const keys = keyGen();
  const msg = "こんにちは";
  const enc = try encrypt(msg, keys.pubKey);
  const dec = try decrypt(enc, keys.pubKey, keys.privKey);
  try std.testing.expectEqualSlices(u8, msg, dec[0..msg.len]);
}