// zig fmt: off
const c = @import("constants.zig");
const bytes = @import("bytes.zig");

fn i2ospLen4(x: u8) [4]u8 {
  var y = x;
  var output = [4]u8{ 0x00, 0x00, 0x00, 0x00 };
  var i: u8 = 0;
  while (y != 0) : (i += 1) {
    const z: u16 = @intCast(y);
    output[i] = @intCast(@mod(z, 256));
    y = @intCast(@divFloor(z, 256));
  }
  return [4]u8{ output[3], output[2], output[1], output[0] };
}

fn makeDbMask(seed: [c.hashOutBytes]u8) [c.dbSize + 1]u8 {
  var output: [c.dbSize + 1]u8 = undefined;
  
  for (0..(c.rsaOutBytes - 2) / c.hashOutBytes) |counter| {
    const x: u8 = @intCast(counter);
    const C = i2ospLen4(x);
    const digest = bytes.hashBytes(seed[0..] ++ C);

    for (counter * c.hashOutBytes..counter * c.hashOutBytes + c.hashOutBytes, 0..c.hashOutBytes) |i, j| {
      output[i] = digest[j];
    }
  }

  return output;
}

pub const PlusEncodeErr = error{ MESSAGE_LONGER_THAN_MAX_MESSAGE_LEN };
pub const PlusDecodeErr = error{ INVALID_LENGTH, INVALID_PADDING };

pub fn plusEncode(m: []const u8) PlusEncodeErr![c.rsaOutBytes]u8 {
  const mLen = m.len;
  if (mLen > c.maxMesLen) {
    return PlusEncodeErr.MESSAGE_LONGER_THAN_MAX_MESSAGE_LEN;
  }

  const hashL = bytes.hashBytes(c.L);
  var db: [c.dbSize]u8 = (hashL ++ ([_]u8{0x00} ** (c.maxMesLen + 1)));
  db[c.dbSize - mLen - 1] = 0x01;

  for ((c.dbSize - mLen)..c.dbSize, 0..mLen) |i, j| {
    db[i] = m[j];
  }

  const seed = bytes.randomBytes();
  const dbMask = makeDbMask(seed)[0..c.dbSize];
  var maskedDb: [c.dbSize]u8 = undefined;

  for (0..c.dbSize) |i| {
    maskedDb[i] = db[i] ^ dbMask[i];
  }

  const seedMask = bytes.hashBytes((maskedDb ++ i2ospLen4(0))[0..(c.dbSize + 4)]);
  var maskedSeed: [c.hashOutBytes]u8 = undefined;

  for (0..c.hashOutBytes) |i| {
    maskedSeed[i] = seed[i] ^ seedMask[i];
  }

  return [1]u8{0x00} ++ maskedSeed ++ maskedDb;
}

pub fn plusDecode(enc: [c.rsaOutBytes]u8) PlusDecodeErr![c.maxMesLen]u8 {
  if (enc.len != c.rsaOutBytes) {
    return PlusDecodeErr.INVALID_LENGTH;
  }

  const maskedSeed = enc[1..c.hashOutBytes + 1];
  const maskedDb = enc[c.hashOutBytes + 1 ..];
  const seedMask = bytes.hashBytes((maskedDb ++ i2ospLen4(0))[0..(c.dbSize + 4)]);
  var seed: [c.hashOutBytes]u8 = undefined;

  for (0..c.hashOutBytes) |i| {
    seed[i] = maskedSeed[i] ^ seedMask[i];
  }

  const dbMask = makeDbMask(seed)[0..c.dbSize];
  var db: [c.dbSize]u8 = undefined;

  for (0..c.dbSize) |i| {
    db[i] = maskedDb[i] ^ dbMask[i];
  }

  const hashL = bytes.hashBytes(c.L);
  for (0..c.hashOutBytes) |i| {
    if (db[i] != hashL[i]) {
      return PlusDecodeErr.INVALID_PADDING;
    }
  }

  var messageStartIdx: u32 = undefined;
  for (c.hashOutBytes..c.dbSize - 1) |i| {
    if (db[i] == 0x01) {
      messageStartIdx = @intCast(i);
      messageStartIdx += 1;
      break;
    }
  }

  var m = [_]u8{0x00} ** c.maxMesLen;
  for (0..c.dbSize - messageStartIdx) |i| {
    m[i] = db[messageStartIdx + i];
  }

  return m;
}