// zig fmt: off
const std = @import("std");
const primes = @import("primes.zig").primes;
const c = @import("constants.zig");
const bytes = @import("bytes.zig");

fn randomHalfRsaOutBitsInt(min: c.halfRsaOutBits, max: c.halfRsaOutBits) c.halfRsaOutBits {
  var out: [c.rsaOutBytes / 2]u8 = undefined;
  var outI: c.halfRsaOutBits = undefined;

  while (true) {
    std.posix.getrandom(out[0..]) catch unreachable;
    outI = @truncate(bytes.toInt(c.halfRsaOutBits, out[0..]));

    if (outI >= min and outI < max) {
      return outI;
    }
  }
}


fn powerMod(x: c.halfRsaOutBits, p: c.halfRsaOutBits, y: c.halfRsaOutBits) c.halfRsaOutBits {
  var res: c.halfRsaOutBits = 1;
  var xV = @mod(x, y);
  var pV = p;
  var temp: c.rsaOutBits = undefined;

  while (pV > 0) {
    if (@mod(pV, 2) != 0) {
        temp = xV;
        temp *= res;
        res = @truncate(@mod(temp, y));
    }

    pV = pV >> 1;
    temp = xV;
    temp *= xV;
    xV = @truncate(@mod(temp, y));
  }

  return res;
}

pub fn powerMod2(x: c.rsaOutBits, p: c.rsaOutBits, y: c.rsaOutBits) c.rsaOutBits {
  var res: c.rsaOutBits = 1;
  var xV = @mod(x, y);
  var pV = p;
  var temp: c.doubleRsaOutBits = undefined;

  while (pV > 0) {
    if (@mod(pV, 2) != 0) {
      temp = xV;
      temp *= res;
      res = @truncate(@mod(temp, y));
    }

    pV = pV >> 1;
    temp = xV;
    temp *= xV;
    xV = @truncate(@mod(temp, y));
  }

  return res;
}

fn millerTest(d: c.halfRsaOutBits, n: c.halfRsaOutBits) bool {
  const a: c.halfRsaOutBits = randomHalfRsaOutBitsInt(2, n - 2);
  var x = powerMod(a, d, n);

  var dV: c.rsaOutBits = d;
  var temp: c.rsaOutBits = undefined;
  if (x == 1 or x == n - 1) return true;

  while (dV != n - 1) {
    temp = x;
    temp *= x;
    x = @truncate(@mod(temp, n));
    if (x == 1) return false;
    if (x == n - 1) return true;
    dV *= 2;
  }

  return false;
}

fn primeTest(n: c.halfRsaOutBits, r: u64) bool {
  if (n <= 1 or n == 4) return false;
  if (n <= 3) return true;
  var d = n - 1;
  while (@mod(d, 2) == 0) d = @divExact(d, 2);
  for (0..r) |i| {
    _ = i;
    if (millerTest(d, n) == false) return false;
  }
  return true;
}

pub fn Gen() c.halfRsaOutBits {
  while (true) {
    var num = randomHalfRsaOutBitsInt(3, c.maxRangeForPrime);
    num = num | (1 << ((c.rsaOutBytes * 4) - 1));
    num = num | (1);
    var divisible = false;

    for (primes) |prime| {
      if (@mod(num, prime) == 0) {
        divisible = true;
        break;
      }
    }

    if (divisible) continue;

    if (primeTest(num, c.primeTestRounds)) {
      return num;
    } else {
      continue;
    }
  }
}

test "powerMod2" {
  const x: c.rsaOutBits = 5;
  const e: c.rsaOutBits = 3;
  const n: c.rsaOutBits = 187;
  const result = powerMod2(x, e, n); // 5^3 mod 187 = 125
  try std.testing.expectEqual(@as(c.rsaOutBits, 125), result);
}